#include <Windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

template<typename T> bool fread_t(FILE*f, T&dst){
	return fread(&dst, 1, sizeof(T), f) == sizeof(T);
}
template<typename T> bool fwrite_t(FILE*f, T&dst){
	return fwrite(&dst, 1, sizeof(T), f) == sizeof(T);
}


#define LZSS_WINDOW_SIZE	4096

#define LZSS_WINDOW_POS(_val) ((_val) % LZSS_WINDOW_SIZE)

enum LZSS_Status {
	LZSS_Success = 0,
	LZSS_ReadError,
	LZSS_WriteError,
};

LZSS_Status LZSSDecompress(FILE* fin, FILE* fout){
	UINT8 flag;
	UINT8 bit;
	UINT8 t, i;
	UINT32 size;
	UINT32 pos, len;
	UINT8 window[LZSS_WINDOW_SIZE];
	memset(window, 0, LZSS_WINDOW_SIZE);

	size = 0xFEE;
	for (;fread_t(fin, flag);){
		t = 0;
		for (bit = 0; bit < 8; bit++, flag >>= 1){
			if (flag & 1){
				if (!fread_t(fin, t))
					return LZSS_Success;
				if (!fwrite_t(fout, t))
					return LZSS_WriteError;
				window[LZSS_WINDOW_POS(size)] = t;
				size++;
				continue;
			}
			if (!fread_t(fin, t))
				return LZSS_Success;
			pos = t;
			if (!fread_t(fin, t))
				return LZSS_ReadError;
			
			len = ((t & 0xF) )+3;
			pos += ((t & 0xF0) << 4);
			pos = pos;
			for (i = 0; i < len; i++){
				if (!fwrite_t(fout, window[LZSS_WINDOW_POS(i+pos)]))
					return LZSS_WriteError;
				window[LZSS_WINDOW_POS(size++)] = window[LZSS_WINDOW_POS(i+pos)];
			}
		}
	}
	return LZSS_Success;
}

int main(int argc, char**argv){
	FILE *fin, *fout;
	if (argc < 3){
		printf("Usage: BmpDecompress <input file> <output file>\n");
		return 0;
	}
	if (fopen_s(&fin, argv[1], "rb") != 0)
		return printf("Failed to open \"%s\" for reading.\n", argv[1]), 0;
	if (fopen_s(&fout, argv[2], "wb") != 0)
		return fclose(fin), printf("Failed to open \"%s\" for writing.\n", argv[2]), 0;
	if (LZSSDecompress(fin, fout) == LZSS_Success)
		printf("Successfully decompressed \"%s\".\n", argv[1]);
	else
		printf("Failed to decompressed \"%s\".\n", argv[1]);
	fclose(fin);
	fclose(fout);
}